##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#                    Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from osv import fields,osv
import ir

class report_custom(osv.osv):
	_name = 'ir.report.custom'
	_columns = {
		'name': fields.char('Report Name',size=64, required=True),
		'groupby': fields.char('Grouped By',size=64),
		'child_ids':  fields.char('Childs fields', size=64),
		'model':  fields.char('Model',size=64, required=True),
		'title': fields.char("Report Header",size=64, required='True'),
		'totalonly': fields.boolean('Show totals only'),
		'print_format':  fields.selection((('a4','a4'),), 'Print format', required=True),
		'print_orientation':  fields.selection((('landscape','Landscape'),('portrait','Portrait')), 'Print orientation', required=True, size=16),
		'repeat_header':  fields.boolean('Repeat Header'),
		'footer':  fields.char('Report Footer', size=64, required=True),
		'sortby':  fields.char('Sorted By', size=64),
		'fields_id': fields.one2many('ir.report.custom.fields', 'report_id','Fields'),
		'state':  fields.selection([('unsubscribed','Unsubscribed'),('subscribed','Subscribed')],'State', size=64),
	}
	_defaults = {
		'print_format': lambda cr,uid,form: 'a4',
		'print_orientation': lambda cr,uid,form: 'portrait',
		'total_only': lambda cr,uid,form: 0,
		'state': lambda cr,uid,form: 'unsubscribed',
		'footer': lambda cr,uid,form: 'Generated by Tiny ERP'
	}
	def subscribe(self, cr, uid, ids, *args):
		for row in self.read(cr, uid, ids, ['model','name']):
			model = row['model']
			name = row['name']
			keys = [('action','client_print_multi'),('res_model',model)]
			result = ir.ir_get(cr, uid, keys)
			for r in result:
				if r[1]==name:
					self.pool.get('ir.actions.report.custom').unlink(cr, uid, [r[2]['id']])
					ir.ir_del(cr, uid, r[0])
			res = {'report_id':row['id'], 'type':'ir.actions.report.custom','name':name}
			id = self.pool.get('ir.actions.report.custom').create(cr, uid, res)
			ir.ir_set(cr, uid, keys, [], name, 'ir.actions.report.custom,'+str(id), isobject=True, replace=False)
		self.write(cr, uid, ids, {'state':'subscribed'})
		return True

report_custom()

class report_custom_fields(osv.osv):
	_name = 'ir.report.custom.fields'
	_columns = {
		'report_id': fields.many2one('ir.report.custom', 'Report Ref'),
		'name': fields.char('Name', size=64, required=True),
		'string': fields.char('String', size=64, required=True),
		'alignment':  fields.selection((('left','left'),('right','right'),('center','center')), 'Alignment', required=True),
		'priority': fields.integer('Priority'),
		'width': fields.integer('Fixed Width'),
		'calc_sum': fields.boolean('Calculate Sum'),
		'calc_avg': fields.boolean('Calculate Average'),
		'calc_count': fields.boolean('Calculate Count'),
		'bgcolor': fields.char('Background Color', size=64),
		'fontcolor': fields.char('Font color', size=64)
	}
	_defaults = {
		'alignment': lambda cr,uid,form: 'left',
		'bgcolor': lambda cr,uid,form: 'white',
		'fontcolor': lambda cr,uid,form: 'black',
		'calc_sum': lambda cr,uid,form: 0,
		'calc_avg': lambda cr,uid,form: 0,
		'calc_count': lambda cr,uid,form: 0,
	}
	_order = "priority"
report_custom_fields()

